import typer
import json
from typing import Annotated

from doipy.actions import create, create_fdo, hello, list_operations, search, get_design, get_init_data
from doipy.server import server

app = typer.Typer()


@app.command(name='create')
def create_command(files: Annotated[list[typer.FileBinaryRead], typer.Argument(help='A list of files for an DO.')],
                   md_file: Annotated[
                       typer.FileText, typer.Option(help='A JSON file containing metadata.')
                   ] = None,
                   client_id: Annotated[str, typer.Option(help='The identifier of the user in Cordra')] = None,
                   password: Annotated[str, typer.Option(help='Password of the user')] = None):
    """Implements 0.DOIP/Op.Create"""
    do_type = 'Document'
    response = create(do_type, files, md_file, client_id, password)
    print(json.dumps(response, indent=2))


@app.command(name='create_fdo')
def create_fdo_command(data_ref: Annotated[str, typer.Argument(help='The identifier of the data object')],
                       md_ref: Annotated[str, typer.Argument(help='The identifier of the metadata object')],
                       client_id: Annotated[str, typer.Option(help='The identifier of the user in Cordra')] = None,
                       password: Annotated[str, typer.Option(help='Password of the user')] = None):
    """Create a new FAIR Digital Object (FDO) from data and metadata references."""
    response = create_fdo(data_ref, md_ref, client_id, password)
    print(json.dumps(response, indent=2))


@app.command(name='hello')
def hello_command(username: Annotated[str, typer.Option(help='username')] = None,
                  password: Annotated[str, typer.Option(help='password')] = None,
                  token: Annotated[str, typer.Option(help='Auth token')] = None):
    """Implements 0.DOIP/Op.Hello: An operation to allow a client to get information about the DOIP service."""
    response = hello(username, password, token)
    print(json.dumps(response, indent=2))


@app.command(name='list_operations')
def list_operations_command(target_id: Annotated[str, typer.Option(help='target_id')] = None,
                            client_id: Annotated[str, typer.Option(help='client_id')] = None,
                            password: Annotated[str, typer.Option(help='password')] = None):
    """
    Implements 0.DOIP/Op.ListOperations: An operation to request the list of operations that can be invoked on the
    target DO.
    """
    response = list_operations(target_id, client_id, password)
    print(json.dumps(response, indent=2))


@app.command(name='search')
def search_command(query: Annotated[str, typer.Argument(help='query')],
                   username: Annotated[str, typer.Option(help='username')] = None,
                   password: Annotated[str, typer.Option(help='password')] = None):
    """Implements 0.DOIP/Op.Search"""
    response = search(query, username, password)
    print(json.dumps(response, indent=2))


@app.command(name='get_design')
def get_design_command():
    """Implements 20.DOIP/Op.GetDesign (see https://www.cordra.org)"""
    response = get_design()
    print(json.dumps(response, indent=2))


@app.command(name='get_init_data')
def get_design_command():
    """Implements 20.DOIP/Op.GetInitData (see https://www.cordra.org)"""
    response = get_init_data()
    print(json.dumps(response, indent=2))


@app.command(name='doip_server')
def doip_server_command(hostname: Annotated[str, typer.Argument(help='hostname')],
                        port: Annotated[int, typer.Argument(help='port')],
                        privkey: Annotated[str, typer.Argument(help='privkey')],
                        cert: Annotated[str, typer.Argument(help='cert')]):
    """Starts a DOIP server"""
    server(hostname, port, privkey, cert)



__all__ = ['create', 'list_operations', 'hello', 'search', 'update', 'create_fdo', 'get_design', 'get_init_data']

from doipy.actions import create, list_operations, hello, search, update, create_fdo, get_design, get_init_data



from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix='doipy_')

# Cordra Testbed Data Repo
    ip: str = '141.5.106.77'
    prefix: str = '21.T11967'
    target_id: str = '21.T11967/service'
    port: int = 9000


settings = Settings()



import json
import uuid
from typing import BinaryIO, TextIO

from rich import print

from doipy.config import settings
from doipy.fdo_config import FDO_Profile_Ref, FDO_Type_Ref
from doipy.socket_utils import create_socket, send_message, finalize_socket, read_response


def create_fdo(data_ref: str, md_ref: str, client_id: str, password: str):
    operation_id = '0.DOIP/Op.Create'
    do_type = 'FDO'
    with create_socket() as ssl_sock:
        message = {
            'clientId': client_id,
            'targetId': f'{settings.target_id}',
            'operationId': operation_id,
            'authentication': {
                'password': password
            }
        }
        send_message(message, ssl_sock)
        message = {
            'type': do_type,
            'attributes': {
                'content': {
                    'id': '',
                    'FDO_Profile_Ref': FDO_Profile_Ref,
                    'FDO_Type_Ref': FDO_Type_Ref,
                    'FDO_Data_Refs': [data_ref],
                    'FDO_MD_Refs': [md_ref]
                }
            }
        }
        send_message(message, ssl_sock)
        finalize_socket(ssl_sock)
        response = read_response(ssl_sock)
        return response


def create(do_type: str, files: list[BinaryIO], md_file: TextIO, client_id: str, password: str):
    operation_id = '0.DOIP/Op.Create'
    if do_type is None:
        return
    with create_socket() as ssl_sock:
        message = {
            'clientId': client_id,
            'targetId': f'{settings.target_id}',
            'operationId': operation_id,
            'authentication': {
                'password': password
            }
        }
        send_message(message, ssl_sock)

        if do_type == 'Document':
            message = {
                'type': do_type,
                'attributes': {
                    'content': {
                        'id': ''
                    }
                }
            }
            my_uuids = {}
            elements = []
            for file in files:
                filename = file.name
                my_uuids[filename] = str(uuid.uuid4())
                elements.append(
                    {
                        'id': my_uuids[filename],
                        'type': 'text/plain',
                        'attributes': {
                            'filename': filename
                        }
                    }
                )
                message['elements'] = elements

            if md_file is not None:
                data = json.load(md_file)
                for key in data:
                    message['attributes']['content'][key] = data[key]
            send_message(message, ssl_sock)

            for filename in my_uuids:
                message = {'id': my_uuids[filename]}
                send_message(message, ssl_sock)

                buffer_size = 1024
                with open(filename, 'rb') as f:
                    while True:
                        bytes_read = f.read(buffer_size)
                        if not bytes_read:
                            break
                        ssl_sock.sendall(bytes_read)

                finalize_socket(ssl_sock)

            finalize_socket(ssl_sock)
            reply = read_response(ssl_sock)
            if reply['status'] == '0.DOIP/Status.102':
                print('Authentication failed')
                return
            else:
                return reply
            # Next step: update the DO to put the correct metadata
            # update(reply['output']['id'], client_id, password, do_type)


def update(target_id: str, client_id: str, password: str, do_type: str):
    operation_id = '0.DOIP/Op.Update'

    with create_socket() as ssl_sock:
        message = {
            'clientId': client_id,
            'targetId': target_id,
            'operationId': operation_id,
            'authentication': {
                'password': password
            }
        }
        send_message(message, ssl_sock)
        string1 = f'https://cordra.testbed.pid.gwdg.de/objects/{target_id}?payload=file'
        string2 = f'https://cordra.testbed.pid.gwdg.de/objects/{target_id}'
        message = {
            'type': do_type,
            'attributes': {
                'content': {
                    'id': '',
                    'Payload': string1,
                    'Metadata': string2
                }
            }
        }
        send_message(message, ssl_sock)
        finalize_socket(ssl_sock)
        response = read_response(ssl_sock)
        return response


def search(query: str = 'type:Document', username: str = None, password: str = None):
    operation_id = '0.DOIP/Op.Search'
    message = {
        'targetId': f'{settings.target_id}',
        'operationId': operation_id,
        'attributes': {
            'query': query
        }
    }
    if username and password:
        message['authentication'] = {'username': username, 'password': password}

    with create_socket() as ssl_sock:
        send_message(message, ssl_sock)
        finalize_socket(ssl_sock)
        response = read_response(ssl_sock)
        return response


def hello(username: str = None, password: str = None, token: str = None):
    operation_id = '0.DOIP/Op.Hello'

    message = {
        'targetId': f'{settings.target_id}',
        'operationId': operation_id,
    }
    if token is not None:
        message['authentication'] = {'token': token}
    elif username is not None:
        if password is None:
            print('Please provide a password!')
            return

    with create_socket() as ssl_sock:
        send_message(message, ssl_sock)
        finalize_socket(ssl_sock)
        response = read_response(ssl_sock)
        return response


def list_operations(target_id: str, client_id: str, password: str):
    operation_id = '0.DOIP/Op.ListOperations'
    message = {'operationId': operation_id}

    if target_id is None:
        message['targetId'] = f'{settings.target_id}'
    else:
        message['targetId'] = target_id

    if client_id is not None and password is not None:
        message['clientId'] = client_id
        message['authentication'] = {'password': password}

    with create_socket() as ssl_sock:
        send_message(message, ssl_sock)
        finalize_socket(ssl_sock)
        response = read_response(ssl_sock)
        return response


def get_design():
    operation_id = '20.DOIP/Op.GetDesign'
    message = {
        "targetId": f'{settings.target_id}',
        "operationId": operation_id
    }
    with create_socket() as ssl_sock:
        send_message(message, ssl_sock)
        finalize_socket(ssl_sock)
        response = read_response(ssl_sock)
        return response


def get_init_data():
    operation_id = '20.DOIP/Op.GetInitData'
    message = {
        "targetId": f'{settings.target_id}',
        "operationId": operation_id
    }
    with create_socket() as ssl_sock:
        send_message(message, ssl_sock)
        finalize_socket(ssl_sock)
        response = read_response(ssl_sock)
        return response



from contextlib import contextmanager
import socket
import ssl
import json

from doipy.config import settings


@contextmanager
def create_socket():
    ssl_settings = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    ssl_settings.check_hostname = False
    ssl_settings.verify_mode = ssl.CERT_NONE
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        with ssl_settings.wrap_socket(sock) as ssl_sock:
            ssl_sock.connect((settings.ip, settings.port))
            yield ssl_sock


def send_message(message, ssl_sock):
    data = (json.dumps(message) + '\n').encode(encoding='UTF-8', errors='ignore')
    ssl_sock.send(data)
    ssl_sock.send(b'#\n')


def send_server_message(message, conn):
    data = (json.dumps(message) + '\n').encode(encoding='UTF-8', errors='ignore')
    print('send now')
    print(data)
    conn.send(data)
    conn.send(b'\n#\n')
    conn.send(b'#\n')


def finalize_socket(ssl_sock):
    ssl_sock.send(b'#\n')


def read_response(ssl_sock):
    not_done = True
    reply_return = ''
    byte_size = 1024
    reply = ssl_sock.recv(byte_size)
    while not_done:
        received = ssl_sock.recv(byte_size)
        if received == b'\n#\n':
            # End of json message

            reply_return = reply.decode()
            reply = ssl_sock.recv(byte_size)
            if reply == b'#\n':
                # End of response
                not_done = False
            #else:
                # Next message starts
        else:
            reply += received
    # fixme: currently only the last full reply is returned
    return json.loads(reply_return)
